@@ -28,10 +28,9 @@ from eventlet import patcher
 from eventlet import tpool
 from oslo_log import log as logging
 from oslo_utils import uuidutils
-import six
 from six.moves import range  # noqa
 
-from os_win._i18n import _, _LE, _LW
+from os_win._i18n import _
 from os_win import _utils
 from os_win import constants
 from os_win import exceptions
@@ -42,6 +41,9 @@ from os_win.utils import pathutils
 
 LOG = logging.getLogger(__name__)
 
+# TODO(claudiub): remove the is_planned_vm argument from methods once it is not
+# used anymore.
+
 
 class VMUtils(baseutils.BaseUtilsVirt):
 
@@ -68,17 +70,21 @@ class VMUtils(baseutils.BaseUtilsVirt):
     _CIM_RES_ALLOC_SETTING_DATA_CLASS = 'Cim_ResourceAllocationSettingData'
     _COMPUTER_SYSTEM_CLASS = "Msvm_ComputerSystem"
     _LOGICAL_IDENTITY_CLASS = 'Msvm_LogicalIdentity'
+    _VIRTUAL_SYSTEM_SNAP_ASSOC_CLASS = 'Msvm_SnapshotOfVirtualSystem'
 
     _S3_DISP_CTRL_RES_SUB_TYPE = 'Microsoft:Hyper-V:S3 Display Controller'
     _SYNTH_DISP_CTRL_RES_SUB_TYPE = ('Microsoft:Hyper-V:Synthetic Display '
                                      'Controller')
-    _SYNTH_3D_DISP_CTRL_RES_SUB_TYPE = ('Microsoft:Hyper-V:Synthetic 3D '
+    _REMOTEFX_DISP_CTRL_RES_SUB_TYPE = ('Microsoft:Hyper-V:Synthetic 3D '
                                         'Display Controller')
-    _SYNTH_3D_DISP_ALLOCATION_SETTING_DATA_CLASS = (
+    _SYNTH_DISP_ALLOCATION_SETTING_DATA_CLASS = (
+        'Msvm_SyntheticDisplayControllerSettingData')
+    _REMOTEFX_DISP_ALLOCATION_SETTING_DATA_CLASS = (
         'Msvm_Synthetic3DDisplayControllerSettingData')
 
     _VIRTUAL_SYSTEM_SUBTYPE = 'VirtualSystemSubType'
     _VIRTUAL_SYSTEM_TYPE_REALIZED = 'Microsoft:Hyper-V:System:Realized'
+    _VIRTUAL_SYSTEM_TYPE_PLANNED = 'Microsoft:Hyper-V:System:Planned'
     _VIRTUAL_SYSTEM_SUBTYPE_GEN2 = 'Microsoft:Hyper-V:SubType:2'
 
     _SNAPSHOT_FULL = 2
@@ -108,23 +114,27 @@ class VMUtils(baseutils.BaseUtilsVirt):
     }
 
     _DISP_CTRL_ADDRESS_DX_11 = "02C1,00000000,01"
+    _DISP_CTRL_ADDRESS = "5353,00000000,00"
 
     _vm_power_states_map = {constants.HYPERV_VM_STATE_ENABLED: 2,
                             constants.HYPERV_VM_STATE_DISABLED: 3,
-                            constants.HYPERV_VM_STATE_SHUTTING_DOWN: 4,
                             constants.HYPERV_VM_STATE_REBOOT: 11,
                             constants.HYPERV_VM_STATE_PAUSED: 9,
                             constants.HYPERV_VM_STATE_SUSPENDED: 6}
 
+    _disk_ctrl_type_mapping = {
+        _SCSI_CTRL_RES_SUB_TYPE: constants.CTRL_TYPE_SCSI,
+        _IDE_CTRL_RES_SUB_TYPE: constants.CTRL_TYPE_IDE
+    }
+
     _DEFAULT_EVENT_CHECK_TIMEFRAME = 60  # seconds
-    _DEFAULT_EVENT_TIMEOUT_MS = 2000
 
     def __init__(self, host='.'):
         super(VMUtils, self).__init__(host)
-        self._jobutils = jobutils.JobUtils()
+        self._jobutils = jobutils.JobUtils(host)
         self._pathutils = pathutils.PathUtils()
         self._enabled_states_map = {v: k for k, v in
-                                    six.iteritems(self._vm_power_states_map)}
+                                    self._vm_power_states_map.items()}
 
     def list_instance_notes(self):
         instance_notes = []
@@ -148,6 +158,8 @@ class VMUtils(baseutils.BaseUtilsVirt):
                     ['ElementName'],
                     VirtualSystemType=self._VIRTUAL_SYSTEM_TYPE_REALIZED)]
 
+    @_utils.not_found_decorator(
+        translated_exc=exceptions.HyperVVMNotFoundException)
     def get_vm_summary_info(self, vm_name):
         vmsettings = self._lookup_vm_check(vm_name)
 
@@ -199,9 +211,10 @@ class VMUtils(baseutils.BaseUtilsVirt):
     def _lookup_vm(self, vm_name, as_vssd=True, for_update=False):
         if as_vssd:
             conn = self._compat_conn if for_update else self._conn
-            vms = conn.Msvm_VirtualSystemSettingData(
-                ElementName=vm_name,
-                VirtualSystemType=self._VIRTUAL_SYSTEM_TYPE_REALIZED)
+            vms = conn.Msvm_VirtualSystemSettingData(ElementName=vm_name)
+            vms = [v for v in vms if
+                   v.VirtualSystemType in [self._VIRTUAL_SYSTEM_TYPE_PLANNED,
+                                           self._VIRTUAL_SYSTEM_TYPE_REALIZED]]
         else:
             vms = self._conn.Msvm_ComputerSystem(ElementName=vm_name)
         n = len(vms)
@@ -214,7 +227,11 @@ class VMUtils(baseutils.BaseUtilsVirt):
             return vms[0]
 
     def vm_exists(self, vm_name):
-        return self._lookup_vm(vm_name) is not None
+        """Checks if the Realized VM with the given name exists on the host."""
+        # NOTE(claudiub): A planned VM and a realized VM cannot exist at the
+        # same time on the same host. The 2 types must be treated separately,
+        # thus, this will only check if the Realized VM exits.
+        return self._lookup_vm(vm_name, False) is not None
 
     def get_vm_id(self, vm_name):
         vm = self._lookup_vm_check(vm_name, as_vssd=False)
@@ -266,56 +283,53 @@ class VMUtils(baseutils.BaseUtilsVirt):
 
         self._jobutils.modify_virt_resource(procsetting)
 
+    def set_nested_virtualization(self, vm_name, state):
+        """Enables nested virtualization for the given VM.
+
+        :raises NotImplemented: Nested virtualization is supported on
+            Windows / Hyper-V Server 2016 or newer.
+        """
+        raise NotImplementedError(_('Nested virtualization is supported on '
+                                    'Windows / Hyper-V Server 2016 or newer.'))
+
     def update_vm(self, vm_name, memory_mb, memory_per_numa_node, vcpus_num,
-                  vcpus_per_numa_node, limit_cpu_features, dynamic_mem_ratio):
-        vmsetting = self._lookup_vm_check(vm_name)
+                  vcpus_per_numa_node, limit_cpu_features, dynamic_mem_ratio,
+                  configuration_root_dir=None, snapshot_dir=None,
+                  host_shutdown_action=None, vnuma_enabled=None,
+                  snapshot_type=None,
+                  is_planned_vm=False):
+        vmsetting = self._lookup_vm_check(vm_name, for_update=True)
+
+        if host_shutdown_action:
+            vmsetting.AutomaticShutdownAction = host_shutdown_action
+        if configuration_root_dir:
+            # Created VMs must have their *DataRoot paths in the same location
+            # as the VM's path.
+            vmsetting.ConfigurationDataRoot = configuration_root_dir
+            vmsetting.LogDataRoot = configuration_root_dir
+            vmsetting.SnapshotDataRoot = configuration_root_dir
+            vmsetting.SuspendDataRoot = configuration_root_dir
+            vmsetting.SwapFileDataRoot = configuration_root_dir
+        if vnuma_enabled is not None:
+            vmsetting.VirtualNumaEnabled = vnuma_enabled
+
         self._set_vm_memory(vmsetting, memory_mb, memory_per_numa_node,
                             dynamic_mem_ratio)
         self._set_vm_vcpus(vmsetting, vcpus_num, vcpus_per_numa_node,
                            limit_cpu_features)
 
+        if snapshot_type:
+            self._set_vm_snapshot_type(vmsetting, snapshot_type)
+
+        self._modify_virtual_system(vmsetting)
+
     def check_admin_permissions(self):
         if not self._compat_conn.Msvm_VirtualSystemManagementService():
             raise exceptions.HyperVAuthorizationException()
 
-    def create_vm(self, *args, **kwargs):
-        # TODO(claudiub): method signature changed. Fix this when the usage
-        # for this method was updated.
-        # update_vm should be called in order to set VM's memory and vCPUs.
-        try:
-            self._create_vm(*args, **kwargs)
-        except TypeError:
-            # the method call was updated to use the _vnuma_create_vm interface
-            self._vnuma_create_vm(*args, **kwargs)
-
-    def _vnuma_create_vm(self, vm_name, vnuma_enabled, vm_gen, instance_path,
-                         notes=None):
+    def create_vm(self, vm_name, vnuma_enabled, vm_gen, instance_path,
+                  notes=None):
         LOG.debug('Creating VM %s', vm_name)
-        self._create_vm_obj(vm_name, vnuma_enabled, vm_gen, notes,
-                            instance_path)
-
-    def _create_vm(self, vm_name, memory_mb, vcpus_num, limit_cpu_features,
-                   dynamic_memory_ratio, vm_gen, instance_path, notes=None):
-        """Creates a VM."""
-
-        LOG.debug('Creating VM %s', vm_name)
-
-        # vNUMA and dynamic memory are mutually exclusive
-        vnuma_enabled = False if dynamic_memory_ratio > 1 else True
-
-        self._create_vm_obj(vm_name, vnuma_enabled, vm_gen, notes,
-                            instance_path)
-
-        vmsetting = self._lookup_vm_check(vm_name)
-
-        LOG.debug('Setting memory for vm %s', vm_name)
-        self._set_vm_memory(vmsetting, memory_mb, None, dynamic_memory_ratio)
-
-        LOG.debug('Set vCPUs for vm %s', vm_name)
-        self._set_vm_vcpus(vmsetting, vcpus_num, None, limit_cpu_features)
-
-    def _create_vm_obj(self, vm_name, vnuma_enabled, vm_gen, notes,
-                       instance_path):
         vs_data = self._compat_conn.Msvm_VirtualSystemSettingData.new()
         vs_data.ElementName = vm_name
         vs_data.Notes = notes
@@ -354,27 +368,47 @@ class VMUtils(baseutils.BaseUtilsVirt):
         return self._get_vm_scsi_controller(vmsettings)
 
     def _get_vm_scsi_controller(self, vmsettings):
+        res = self._get_vm_disk_controllers(vmsettings,
+                                            self._SCSI_CTRL_RES_SUB_TYPE)
+        return res[0].path_() if res else None
+
+    def _get_vm_disk_controllers(self, vmsettings, ctrl_res_sub_type):
         rasds = _wqlutils.get_element_associated_class(
             self._conn, self._RESOURCE_ALLOC_SETTING_DATA_CLASS,
             element_instance_id=vmsettings.InstanceID)
         res = [r for r in rasds
-               if r.ResourceSubType == self._SCSI_CTRL_RES_SUB_TYPE][0]
-        return res.path_()
+               if r.ResourceSubType == ctrl_res_sub_type]
+        return res
 
     def _get_vm_ide_controller(self, vmsettings, ctrller_addr):
-        rasds = _wqlutils.get_element_associated_class(
-            self._conn, self._RESOURCE_ALLOC_SETTING_DATA_CLASS,
-            element_instance_id=vmsettings.InstanceID)
-        ide_ctrls = [r for r in rasds
-                     if r.ResourceSubType == self._IDE_CTRL_RES_SUB_TYPE and
-                     r.Address == str(ctrller_addr)]
+        ide_ctrls = self._get_vm_disk_controllers(vmsettings,
+                                                  self._IDE_CTRL_RES_SUB_TYPE)
+        ctrl = [r for r in ide_ctrls
+                if r.Address == str(ctrller_addr)]
 
-        return ide_ctrls[0].path_() if ide_ctrls else None
+        return ctrl[0].path_() if ctrl else None
 
     def get_vm_ide_controller(self, vm_name, ctrller_addr):
         vmsettings = self._lookup_vm_check(vm_name)
         return self._get_vm_ide_controller(vmsettings, ctrller_addr)
 
+    def _get_disk_ctrl_addr(self, controller_path):
+        ctrl = self._get_wmi_obj(controller_path)
+        if ctrl.ResourceSubType == self._IDE_CTRL_RES_SUB_TYPE:
+            return ctrl.Address
+
+        vmsettings = ctrl.associators(
+            wmi_result_class=self._VIRTUAL_SYSTEM_SETTING_DATA_CLASS)[0]
+        # The powershell commandlets rely on the controller index as SCSI
+        # controllers are missing the 'Address' attribute. We'll do the
+        # same.
+        scsi_ctrls = self._get_vm_disk_controllers(
+            vmsettings, self._SCSI_CTRL_RES_SUB_TYPE)
+        ctrl_paths = [rasd.path_().upper() for rasd in scsi_ctrls]
+
+        if controller_path.upper() in ctrl_paths:
+            return ctrl_paths.index(controller_path.upper())
+
     def get_attached_disks(self, scsi_controller_path):
         volumes = self._conn.query(
             self._get_attached_disks_query_string(scsi_controller_path))
@@ -453,8 +487,49 @@ class VMUtils(baseutils.BaseUtilsVirt):
         res.Parent = drive_path
         res.HostResource = [path]
 
-        # Add the new vhd object as a virtual hard disk to the vm.
-        self._jobutils.add_virt_resource(res, vm)
+        try:
+            # Add the new vhd object as a virtual hard disk to the vm.
+            self._jobutils.add_virt_resource(res, vm)
+        except Exception:
+            LOG.exception("Failed to attach disk image %(disk_path)s "
+                          "to vm %(vm_name)s. Reverting attachment.",
+                          dict(disk_path=path, vm_name=vm_name))
+
+            drive = self._get_wmi_obj(drive_path)
+            self._jobutils.remove_virt_resource(drive)
+            raise
+
+    def get_disk_attachment_info(self, attached_disk_path=None,
+                                 is_physical=True, serial=None):
+        res = self._get_mounted_disk_resource_from_path(attached_disk_path,
+                                                        is_physical,
+                                                        serial=serial)
+        if not res:
+            err_msg = _("Disk '%s' is not attached to a vm.")
+            raise exceptions.DiskNotFound(err_msg % attached_disk_path)
+
+        if is_physical:
+            drive = res
+        else:
+            drive = self._get_wmi_obj(res.Parent)
+
+        ctrl_slot = int(drive.AddressOnParent)
+        ctrl_path = drive.Parent
+        ctrl_type = self._get_disk_controller_type(ctrl_path)
+        ctrl_addr = self._get_disk_ctrl_addr(ctrl_path)
+
+        attachment_info = dict(controller_slot=ctrl_slot,
+                               controller_path=ctrl_path,
+                               controller_type=ctrl_type,
+                               controller_addr=ctrl_addr)
+        return attachment_info
+
+    def _get_disk_controller_type(self, controller_path):
+        ctrl = self._get_wmi_obj(controller_path)
+        res_sub_type = ctrl.ResourceSubType
+
+        ctrl_type = self._disk_ctrl_type_mapping[res_sub_type]
+        return ctrl_type
 
     def create_scsi_controller(self, vm_name):
         """Create an iscsi controller ready to mount volumes."""
@@ -488,9 +563,10 @@ class VMUtils(baseutils.BaseUtilsVirt):
             diskdrive.ElementName = serial
             self._jobutils.modify_virt_resource(diskdrive)
 
-    def get_vm_physical_disk_mapping(self, vm_name):
+    def get_vm_physical_disk_mapping(self, vm_name, is_planned_vm=False):
         mapping = {}
-        physical_disks = self.get_vm_disks(vm_name)[1]
+        physical_disks = (
+            self.get_vm_disks(vm_name)[1])
         for diskdrive in physical_disks:
             mapping[diskdrive.ElementName] = dict(
                 resource_path=diskdrive.path_(),
@@ -505,49 +581,32 @@ class VMUtils(baseutils.BaseUtilsVirt):
         diskdrive.HostResource = [mounted_disk_path]
         self._jobutils.modify_virt_resource(diskdrive)
 
-    def set_disk_host_resource(self, vm_name, controller_path, address,
-                               mounted_disk_path):
-        # TODO(lpetrut): remove this method after the patch fixing
-        # swapped disks after host reboot merges in Nova.
-        disk_found = False
-        vmsettings = self._lookup_vm_check(vm_name)
-        (disk_resources, volume_resources) = self._get_vm_disks(vmsettings)
-        for disk_resource in disk_resources + volume_resources:
-            if (disk_resource.Parent == controller_path and
-                    self._get_disk_resource_address(disk_resource) ==
-                    str(address)):
-                if (disk_resource.HostResource and
-                        disk_resource.HostResource[0] != mounted_disk_path):
-                    LOG.debug('Updating disk host resource "%(old)s" to '
-                              '"%(new)s"' %
-                              {'old': disk_resource.HostResource[0],
-                               'new': mounted_disk_path})
-                    disk_resource.HostResource = [mounted_disk_path]
-                    self._jobutils.modify_virt_resource(disk_resource)
-                disk_found = True
-                break
-        if not disk_found:
-            LOG.warning(_LW('Disk not found on controller '
-                            '"%(controller_path)s" with '
-                            'address "%(address)s"'),
-                        {'controller_path': controller_path,
-                         'address': address})
-
     def _get_nic_data_by_name(self, name):
-        return self._conn.Msvm_SyntheticEthernetPortSettingData(
-            ElementName=name)[0]
+        nics = self._conn.Msvm_SyntheticEthernetPortSettingData(
+            ElementName=name)
+        if nics:
+            return nics[0]
 
-    def create_nic(self, vm_name, nic_name, mac_address):
-        """Create a (synthetic) nic and attach it to the vm."""
+        raise exceptions.HyperVvNicNotFound(vnic_name=name)
+
+    def create_nic(self, vm_name, nic_name, mac_address=None):
+        """Create a (synthetic) nic and attach it to the vm.
+
+        :param vm_name: The VM name to which the NIC will be attached to.
+        :param nic_name: The name of the NIC to be attached.
+        :param mac_address: The VM NIC's MAC address. If None, a Dynamic MAC
+            address will be used instead.
+        """
         # Create a new nic
         new_nic_data = self._get_new_setting_data(
             self._SYNTHETIC_ETHERNET_PORT_SETTING_DATA_CLASS)
 
         # Configure the nic
         new_nic_data.ElementName = nic_name
-        new_nic_data.Address = mac_address.replace(':', '')
-        new_nic_data.StaticMacAddress = 'True'
         new_nic_data.VirtualSystemIdentifiers = ['{' + str(uuid.uuid4()) + '}']
+        if mac_address:
+            new_nic_data.Address = mac_address.replace(':', '')
+            new_nic_data.StaticMacAddress = 'True'
 
         # Add the new nic to the vm
         vmsettings = self._lookup_vm_check(vm_name)
@@ -561,8 +620,13 @@ class VMUtils(baseutils.BaseUtilsVirt):
         :param nic_name: The NIC's ElementName.
         """
         # TODO(claudiub): remove vm_name argument, no longer used.
-        nic_data = self._get_nic_data_by_name(nic_name)
-        self._jobutils.remove_virt_resource(nic_data)
+        try:
+            nic_data = self._get_nic_data_by_name(nic_name)
+            self._jobutils.remove_virt_resource(nic_data)
+        except exceptions.NotFound:
+            LOG.debug("Ignoring NotFound exception while attempting "
+                      "to remove vm nic: '%s'. It may have been already "
+                      "deleted.", nic_name)
 
     def soft_shutdown_vm(self, vm_name):
         vm = self._lookup_vm_check(vm_name, as_vssd=False)
@@ -579,6 +643,7 @@ class VMUtils(baseutils.BaseUtilsVirt):
                                                              Reason=reason)
         self._jobutils.check_ret_val(ret_val, None)
 
+    @_utils.retry_decorator(exceptions=exceptions.WMIJobFailed)
     def set_vm_state(self, vm_name, req_state):
         """Set the desired state of the VM."""
 
@@ -595,7 +660,11 @@ class VMUtils(baseutils.BaseUtilsVirt):
     def _get_disk_resource_disk_path(self, disk_resource):
         return disk_resource.HostResource
 
-    def get_vm_storage_paths(self, vm_name):
+    def get_vm_config_root_dir(self, vm_name):
+        vmsettings = self._lookup_vm_check(vm_name)
+        return vmsettings.ConfigurationDataRoot
+
+    def get_vm_storage_paths(self, vm_name, is_planned_vm=False):
         vmsettings = self._lookup_vm_check(vm_name)
         (disk_resources, volume_resources) = self._get_vm_disks(vmsettings)
 
@@ -611,7 +680,7 @@ class VMUtils(baseutils.BaseUtilsVirt):
 
         return (disk_files, volume_drives)
 
-    def get_vm_disks(self, vm_name):
+    def get_vm_disks(self, vm_name, is_planned_vm=False):
         vmsettings = self._lookup_vm_check(vm_name)
         return self._get_vm_disks(vmsettings)
 
@@ -642,21 +711,34 @@ class VMUtils(baseutils.BaseUtilsVirt):
         (job_path, ret_val) = self._vs_man_svc.DestroySystem(vm.path_())
         self._jobutils.check_ret_val(ret_val, job_path)
 
-    def take_vm_snapshot(self, vm_name):
+    def take_vm_snapshot(self, vm_name, snapshot_name=None):
         vm = self._lookup_vm_check(vm_name, as_vssd=False)
         vs_snap_svc = self._compat_conn.Msvm_VirtualSystemSnapshotService()[0]
 
         (job_path, snp_setting_data, ret_val) = vs_snap_svc.CreateSnapshot(
             AffectedSystem=vm.path_(),
             SnapshotType=self._SNAPSHOT_FULL)
-        self._jobutils.check_ret_val(ret_val, job_path)
 
-        vm_path = vm.path_().lower()
-        current_snapshots = self._conn.Msvm_MostCurrentSnapshotInBranch()
-        snp_setting_data = [s for s in current_snapshots if
-                            s.Antecedent.path_().lower() == vm_path][0]
+        job = self._jobutils.check_ret_val(ret_val, job_path)
+        snp_setting_data = job.associators(
+            wmi_result_class=self._VIRTUAL_SYSTEM_SETTING_DATA_CLASS,
+            wmi_association_class=self._AFFECTED_JOB_ELEMENT_CLASS)[0]
 
-        return snp_setting_data.Dependent.path_()
+        if snapshot_name is not None:
+            snp_setting_data.ElementName = snapshot_name
+            self._modify_virtual_system(snp_setting_data)
+
+        return snp_setting_data.path_()
+
+    def get_vm_snapshots(self, vm_name, snapshot_name=None):
+        vm = self._lookup_vm_check(vm_name, as_vssd=False)
+        snapshots = vm.associators(
+            wmi_association_class=self._VIRTUAL_SYSTEM_SNAP_ASSOC_CLASS,
+            wmi_result_class=self._VIRTUAL_SYSTEM_SETTING_DATA_CLASS)
+
+        return [
+            s.path_() for s in snapshots
+            if snapshot_name is None or s.ElementName == snapshot_name]
 
     def remove_vm_snapshot(self, snapshot_path):
         vs_snap_svc = self._compat_conn.Msvm_VirtualSystemSnapshotService()[0]
@@ -680,10 +762,11 @@ class VMUtils(baseutils.BaseUtilsVirt):
                                                                   is_physical)
         return disk_resource is not None
 
-    def detach_vm_disk(self, vm_name, disk_path, is_physical=True):
+    def detach_vm_disk(self, vm_name, disk_path=None, is_physical=True,
+                       serial=None):
         # TODO(claudiub): remove vm_name argument, no longer used.
-        disk_resource = self._get_mounted_disk_resource_from_path(disk_path,
-                                                                  is_physical)
+        disk_resource = self._get_mounted_disk_resource_from_path(
+            disk_path, is_physical, serial=serial)
 
         if disk_resource:
             parent = self._conn.query("SELECT * FROM "
@@ -695,7 +778,8 @@ class VMUtils(baseutils.BaseUtilsVirt):
             if not is_physical:
                 self._jobutils.remove_virt_resource(parent)
 
-    def _get_mounted_disk_resource_from_path(self, disk_path, is_physical):
+    def _get_mounted_disk_resource_from_path(self, disk_path, is_physical,
+                                             serial=None):
         if is_physical:
             class_name = self._RESOURCE_ALLOC_SETTING_DATA_CLASS
         else:
@@ -710,9 +794,15 @@ class VMUtils(baseutils.BaseUtilsVirt):
                      'res_sub_type_virt': self._HARD_DISK_RES_SUB_TYPE,
                      'res_sub_type_dvd': self._DVD_DISK_RES_SUB_TYPE})
 
+        if serial:
+            query += " AND ElementName='%s'" % serial
+
         disk_resources = self._compat_conn.query(query)
 
         for disk_resource in disk_resources:
+            if serial:
+                return disk_resource
+
             if disk_resource.HostResource:
                 if disk_resource.HostResource[0].lower() == disk_path.lower():
                     return disk_resource
@@ -750,25 +840,6 @@ class VMUtils(baseutils.BaseUtilsVirt):
         raise exceptions.HyperVException(
             _("Exceeded the maximum number of slots"))
 
-    def get_vm_serial_port_connection(self, vm_name, update_connection=None):
-        # TODO(lpetrut): Remove this method after the patch implementing
-        # serial console access support merges in Nova.
-        vmsettings = self._lookup_vm_check(vm_name)
-
-        rasds = _wqlutils.get_element_associated_class(
-            self._compat_conn, self._SERIAL_PORT_SETTING_DATA_CLASS,
-            element_instance_id=vmsettings.InstanceID)
-        serial_port = (
-            [r for r in rasds if
-             r.ResourceSubType == self._SERIAL_PORT_RES_SUB_TYPE][0])
-
-        if update_connection:
-            serial_port.Connection = [update_connection]
-            self._jobutils.modify_virt_resource(serial_port)
-
-        if len(serial_port.Connection) > 0:
-            return serial_port.Connection[0]
-
     def _get_vm_serial_ports(self, vmsettings):
         rasds = _wqlutils.get_element_associated_class(
             self._compat_conn, self._SERIAL_PORT_SETTING_DATA_CLASS,
@@ -808,7 +879,7 @@ class VMUtils(baseutils.BaseUtilsVirt):
 
     def get_vm_power_state_change_listener(
             self, timeframe=_DEFAULT_EVENT_CHECK_TIMEFRAME,
-            event_timeout=_DEFAULT_EVENT_TIMEOUT_MS,
+            event_timeout=constants.DEFAULT_WMI_EVENT_TIMEOUT_MS,
             filtered_states=None, get_handler=False):
         field = self._VM_ENABLED_STATE_PROP
         query = self._get_event_wql_query(cls=self._COMPUTER_SYSTEM_CLASS,
@@ -843,10 +914,10 @@ class VMUtils(baseutils.BaseUtilsVirt):
                     try:
                         callback(vm_name, vm_power_state)
                     except Exception:
-                        err_msg = _LE("Executing VM power state change event "
-                                      "callback failed. "
-                                      "VM name: %(vm_name)s, "
-                                      "VM power state: %(vm_power_state)s.")
+                        err_msg = ("Executing VM power state change "
+                                   "event callback failed. "
+                                   "VM name: %(vm_name)s, "
+                                   "VM power state: %(vm_power_state)s.")
                         LOG.exception(err_msg,
                                       dict(vm_name=vm_name,
                                            vm_power_state=vm_power_state))
@@ -854,8 +925,8 @@ class VMUtils(baseutils.BaseUtilsVirt):
                     pass
                 except Exception:
                     LOG.exception(
-                        _LE("The VM power state change event listener "
-                            "encountered an unexpected exception."))
+                        "The VM power state change event listener "
+                        "encountered an unexpected exception.")
                     time.sleep(event_timeout / 1000)
 
         return _handle_events if get_handler else listener
@@ -941,28 +1012,30 @@ class VMUtils(baseutils.BaseUtilsVirt):
                   'this Hyper-V version.'))
 
     def set_disk_qos_specs(self, disk_path, max_iops=None, min_iops=None):
-        if min_iops is None and max_iops is None:
-            LOG.debug("Skipping setting disk QoS specs as no "
-                      "value was provided.")
-            return
+        """Hyper-V disk QoS policy.
 
-        disk_resource = self._get_mounted_disk_resource_from_path(
-            disk_path, is_physical=False)
+        This feature is supported on Windows / Hyper-V Server 2012 R2 or newer.
 
-        if max_iops is not None:
-            disk_resource.IOPSLimit = max_iops
-        if min_iops is not None:
-            disk_resource.IOPSReservation = min_iops
-
-        self._jobutils.modify_virt_resource(disk_resource)
-
-    def _is_drive_physical(self, drive_path):
-        # TODO(atuvenie): Find better way to check if path represents
-        # physical or virtual drive.
-        return not self._pathutils.exists(drive_path)
+        :raises os_win.exceptions.UnsupportedOperation: if the given max_iops
+            or min_iops have non-zero values.
+        """
+        if min_iops or max_iops:
+            raise exceptions.UnsupportedOperation(
+                reason=_("Virtual disk QoS is not supported on this "
+                         "hypervisor version."))
 
     def _drive_to_boot_source(self, drive_path):
-        is_physical = self._is_drive_physical(drive_path)
+        # We expect the drive path to be the one that was passed to the
+        # 'attach_drive' or 'attach_volume_to_controller' methods. In case of
+        # passthrough disks, the drive path will be a Msvm_DiskDrive WMI
+        # object path while for image files it will be the actual image path.
+        #
+        # Note that Msvm_DiskDrive objects will also exist for attached disk
+        # images, but that's not what we'll get in this situation. If we ever
+        # need to accept Msvm_DiskDrive object paths for image files as well,
+        # an extra check will be needed, but that may lead to some other
+        # inconsistencies.
+        is_physical = r'root\virtualization\v2:Msvm_DiskDrive' in drive_path
         drive = self._get_mounted_disk_resource_from_path(
             drive_path, is_physical=is_physical)
 
@@ -993,7 +1066,7 @@ class VMUtils(baseutils.BaseUtilsVirt):
 
         # NOTE(abalutoiu): new_boot_order will contain ROOT uppercase
         # in the device paths while old_boot_order will contain root
-        # lowercase, which will cause the tupple addition result to contain
+        # lowercase, which will cause the tuple addition result to contain
         # each device path twice because of the root lowercase and uppercase.
         # Forcing all the device paths to uppercase fixes the issue.
         new_boot_order = [x.upper() for x in new_boot_order]
@@ -1029,44 +1102,86 @@ class VMUtils(baseutils.BaseUtilsVirt):
                    'max_monitors':
                    self._remotefx_max_monitors_map[max_resolution]})
 
-    def _add_3d_display_controller(self, vm, monitor_count,
-                                   max_resolution, vram_bytes=None):
-        synth_3d_disp_ctrl_res = self._get_new_resource_setting_data(
-            self._SYNTH_3D_DISP_CTRL_RES_SUB_TYPE,
-            self._SYNTH_3D_DISP_ALLOCATION_SETTING_DATA_CLASS)
+    def _set_remotefx_display_controller(self, vm, remotefx_disp_ctrl_res,
+                                         monitor_count, max_resolution,
+                                         vram_bytes=None):
+        new_wmi_obj = False
+        if not remotefx_disp_ctrl_res:
+            new_wmi_obj = True
+            remotefx_disp_ctrl_res = self._get_new_resource_setting_data(
+                self._REMOTEFX_DISP_CTRL_RES_SUB_TYPE,
+                self._REMOTEFX_DISP_ALLOCATION_SETTING_DATA_CLASS)
 
-        synth_3d_disp_ctrl_res.MaximumMonitors = monitor_count
-        synth_3d_disp_ctrl_res.MaximumScreenResolution = max_resolution
+        remotefx_disp_ctrl_res.MaximumMonitors = monitor_count
+        remotefx_disp_ctrl_res.MaximumScreenResolution = max_resolution
+        self._set_remotefx_vram(remotefx_disp_ctrl_res, vram_bytes)
 
-        self._jobutils.add_virt_resource(synth_3d_disp_ctrl_res, vm)
+        if new_wmi_obj:
+            self._jobutils.add_virt_resource(remotefx_disp_ctrl_res, vm)
+        else:
+            self._jobutils.modify_virt_resource(remotefx_disp_ctrl_res)
+
+    def _set_remotefx_vram(self, remotefx_disp_ctrl_res, vram_bytes):
+        pass
 
     def enable_remotefx_video_adapter(self, vm_name, monitor_count,
                                       max_resolution, vram_bytes=None):
-        vm = self._lookup_vm_check(vm_name, as_vssd=False)
-
         self._validate_remotefx_params(monitor_count, max_resolution,
                                        vram_bytes=vram_bytes)
 
+        vm = self._lookup_vm_check(vm_name)
         rasds = _wqlutils.get_element_associated_class(
             self._compat_conn, self._CIM_RES_ALLOC_SETTING_DATA_CLASS,
-            element_uuid=vm.Name)
-        if [r for r in rasds if r.ResourceSubType ==
-                self._SYNTH_3D_DISP_CTRL_RES_SUB_TYPE]:
-            raise exceptions.HyperVRemoteFXException(
-                _("RemoteFX is already configured for this VM"))
+            element_instance_id=vm.InstanceID)
 
         synth_disp_ctrl_res_list = [r for r in rasds if r.ResourceSubType ==
                                     self._SYNTH_DISP_CTRL_RES_SUB_TYPE]
         if synth_disp_ctrl_res_list:
+            # we need to remove the generic display controller first.
             self._jobutils.remove_virt_resource(synth_disp_ctrl_res_list[0])
 
+        remotefx_disp_ctrl_res = [r for r in rasds if r.ResourceSubType ==
+                                  self._REMOTEFX_DISP_CTRL_RES_SUB_TYPE]
+        remotefx_disp_ctrl_res = (remotefx_disp_ctrl_res[0]
+                                  if remotefx_disp_ctrl_res else None)
+
         max_res_value = self._remote_fx_res_map.get(max_resolution)
-        self._add_3d_display_controller(vm, monitor_count, max_res_value,
-                                        vram_bytes)
-        if self._vm_has_s3_controller(vm.ElementName):
+        self._set_remotefx_display_controller(
+            vm, remotefx_disp_ctrl_res, monitor_count, max_res_value,
+            vram_bytes)
+
+        if self._vm_has_s3_controller(vm_name):
             s3_disp_ctrl_res = [r for r in rasds if r.ResourceSubType ==
                                 self._S3_DISP_CTRL_RES_SUB_TYPE][0]
-            s3_disp_ctrl_res.Address = self._DISP_CTRL_ADDRESS_DX_11
+            if s3_disp_ctrl_res.Address != self._DISP_CTRL_ADDRESS_DX_11:
+                s3_disp_ctrl_res.Address = self._DISP_CTRL_ADDRESS_DX_11
+                self._jobutils.modify_virt_resource(s3_disp_ctrl_res)
+
+    def disable_remotefx_video_adapter(self, vm_name):
+        vm = self._lookup_vm_check(vm_name)
+        rasds = _wqlutils.get_element_associated_class(
+            self._compat_conn, self._CIM_RES_ALLOC_SETTING_DATA_CLASS,
+            element_instance_id=vm.InstanceID)
+
+        remotefx_disp_ctrl_res = [r for r in rasds if r.ResourceSubType ==
+                                  self._REMOTEFX_DISP_CTRL_RES_SUB_TYPE]
+
+        if not remotefx_disp_ctrl_res:
+            # VM does not have RemoteFX configured.
+            return
+
+        # we need to remove the RemoteFX display controller first.
+        self._jobutils.remove_virt_resource(remotefx_disp_ctrl_res[0])
+
+        synth_disp_ctrl_res = self._get_new_resource_setting_data(
+            self._SYNTH_DISP_CTRL_RES_SUB_TYPE,
+            self._SYNTH_DISP_ALLOCATION_SETTING_DATA_CLASS)
+        self._jobutils.add_virt_resource(synth_disp_ctrl_res, vm)
+
+        if self._vm_has_s3_controller(vm_name):
+            s3_disp_ctrl_res = [r for r in rasds if r.ResourceSubType ==
+                                self._S3_DISP_CTRL_RES_SUB_TYPE][0]
+            s3_disp_ctrl_res.Address = self._DISP_CTRL_ADDRESS
             self._jobutils.modify_virt_resource(s3_disp_ctrl_res)
 
     def _vm_has_s3_controller(self, vm_name):
@@ -1074,3 +1189,85 @@ class VMUtils(baseutils.BaseUtilsVirt):
 
     def is_secure_vm(self, instance_name):
         return False
+
+    def update_vm_disk_path(self, disk_path, new_disk_path, is_physical=True):
+        disk_resource = self._get_mounted_disk_resource_from_path(
+            disk_path=disk_path, is_physical=is_physical)
+        disk_resource.HostResource = [new_disk_path]
+        self._jobutils.modify_virt_resource(disk_resource)
+
+    def add_pci_device(self, vm_name, vendor_id, product_id):
+        """Adds the given PCI device to the given VM.
+
+        :raises NotImplemented: PCI passthrough is supported on
+            Windows / Hyper-V Server 2016 or newer.
+        """
+        raise NotImplementedError(_('PCI passthrough is supported on '
+                                    'Windows / Hyper-V Server 2016 or newer.'))
+
+    def remove_pci_device(self, vm_name, vendor_id, product_id):
+        """Removes the given PCI device from the given VM.
+
+        :raises NotImplemented: PCI passthrough is supported on
+            Windows / Hyper-V Server 2016 or newer.
+        """
+        raise NotImplementedError(_('PCI passthrough is supported on '
+                                    'Windows / Hyper-V Server 2016 or newer.'))
+
+    def remove_all_pci_devices(self, vm_name):
+        """Removes all the PCI devices from the given VM.
+
+        There are no PCI devices attached to Windows / Hyper-V Server 2012 R2
+        or older VMs.
+        """
+
+    def _set_vm_snapshot_type(self, vmsettings, snapshot_type):
+        # Supported on Windows Server 2016 or newer.
+        pass
+
+    def populate_fsk(self, fsk_filepath, fsk_pairs):
+        """Writes the given FSK pairs into the give file.
+
+        :raises NotImplementedError: This method is required for Shielded VMs,
+            which is supported on Windows / Hyper-V Server 2016 or newer.
+        """
+        raise NotImplementedError(_('This method is supported on Windows / '
+                                    'Hyper-V Server 2016 or newer'))
+
+    def add_vtpm(self, vm_name, pdk_filepath, shielded):
+        """Adds a vtpm and enables it with encryption or shielded option.
+
+        :raises NotImplementedError: This method is required for Shielded VMs,
+            which is supported on Windows / Hyper-V Server 2016 or newer.
+        """
+        raise NotImplementedError(_('This method is supported on Windows / '
+                                    'Hyper-V Server 2016 or newer'))
+
+    def provision_vm(self, vm_name, fsk_filepath, pdk_filepath):
+        """Provisions the given VM with the given FSK and PDK files.
+
+        :raises NotImplementedError: This method is required for Shielded VMs,
+            which is supported on Windows / Hyper-V Server 2016 or newer.
+        """
+        raise NotImplementedError(_('This method is supported on Windows / '
+                                    'Hyper-V Server 2016 or newer'))
+
+
+class VMUtils6_3(VMUtils):
+
+    def set_disk_qos_specs(self, disk_path, max_iops=None, min_iops=None):
+        """Sets the disk's QoS policy."""
+        if min_iops is None and max_iops is None:
+            LOG.debug("Skipping setting disk QoS specs as no "
+                      "value was provided.")
+            return
+
+        disk_resource = self._get_mounted_disk_resource_from_path(
+            disk_path, is_physical=False)
+
+        if max_iops is not None:
+            disk_resource.IOPSLimit = max_iops
+        if min_iops is not None:
+            disk_resource.IOPSReservation = min_iops
+
+        self._jobutils.modify_virt_resource(disk_resource)
