@@ -15,28 +15,38 @@
 #    under the License.
 
 import ctypes
-import sys
 
 from oslo_log import log as logging
 
+from os_win import _utils
 from os_win import exceptions
+from os_win.utils.winapi import constants as w_const
+from os_win.utils.winapi import libs as w_lib
+from os_win.utils.winapi import wintypes
 
-if sys.platform == 'win32':
-    kernel32 = ctypes.windll.kernel32
+kernel32 = w_lib.get_shared_lib_handle(w_lib.KERNEL32)
 
 LOG = logging.getLogger(__name__)
 
-FORMAT_MESSAGE_FROM_SYSTEM = 0x00001000
-FORMAT_MESSAGE_ALLOCATE_BUFFER = 0x00000100
-FORMAT_MESSAGE_IGNORE_INSERTS = 0x00000200
-
 
 class Win32Utils(object):
     def __init__(self):
         self._kernel32_lib_func_opts = dict(error_on_nonzero_ret_val=False,
                                             ret_val_is_err_code=False)
 
-    def run_and_check_output(self, func, *args, **kwargs):
+    def run_and_check_output(self, *args, **kwargs):
+        eventlet_nonblocking_mode = kwargs.pop(
+            'eventlet_nonblocking_mode', True)
+
+        if eventlet_nonblocking_mode:
+            # We have to make sure that the invoked function as well as the
+            # subsequent error handling are performed within the same thread.
+            return _utils.avoid_blocking_call(
+                self._run_and_check_output, *args, **kwargs)
+        else:
+            return self._run_and_check_output(*args, **kwargs)
+
+    def _run_and_check_output(self, func, *args, **kwargs):
         """Convenience helper method for running Win32 API methods."""
         kernel32_lib_func = kwargs.pop('kernel32_lib_func', False)
         if kernel32_lib_func:
@@ -86,8 +96,9 @@ class Win32Utils(object):
         message_buffer = ctypes.c_char_p()
 
         kernel32.FormatMessageA(
-            FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_ALLOCATE_BUFFER |
-            FORMAT_MESSAGE_IGNORE_INSERTS,
+            w_const.FORMAT_MESSAGE_FROM_SYSTEM |
+            w_const.FORMAT_MESSAGE_ALLOCATE_BUFFER |
+            w_const.FORMAT_MESSAGE_IGNORE_INSERTS,
             None, error_code, 0, ctypes.byref(message_buffer), 0, None)
 
         error_message = message_buffer.value
@@ -105,13 +116,32 @@ class Win32Utils(object):
         return hresult & 0xFFFF
 
     def get_com_err_code(self, com_error):
-        hres = self.get_com_error_hresult(com_error)
+        hres = _utils.get_com_error_hresult(com_error)
         if hres is not None:
             return self.hresult_to_err_code(hres)
 
-    @staticmethod
-    def get_com_error_hresult(com_error):
+    def local_free(self, handle):
         try:
-            return ctypes.c_uint(com_error.excepinfo[5]).value
-        except Exception:
-            LOG.debug("Unable to retrieve COM error hresult: %s", com_error)
+            self._run_and_check_output(kernel32.LocalFree, handle)
+        except exceptions.Win32Exception:
+            LOG.exception("Could not deallocate memory. "
+                          "There could be a memory leak.")
+
+    def close_handle(self, handle):
+        kernel32.CloseHandle(handle)
+
+    def wait_for_multiple_objects(self, handles, wait_all=True,
+                                  milliseconds=w_const.INFINITE):
+        handle_array = (wintypes.HANDLE * len(handles))(*handles)
+        ret_val = self.run_and_check_output(
+            kernel32.WaitForMultipleObjects,
+            len(handles),
+            handle_array,
+            wait_all,
+            milliseconds,
+            kernel32_lib_func=True,
+            error_ret_vals=[w_const.WAIT_FAILED])
+        if ret_val == w_const.ERROR_WAIT_TIMEOUT:
+            raise exceptions.Timeout()
+
+        return ret_val
