@@ -13,24 +13,12 @@
 #    License for the specific language governing permissions and limitations
 #    under the License.
 
-from oslo_config import cfg
 from oslo_utils import importutils
 
-from os_win._i18n import _, _LW  # noqa
+from os_win._i18n import _  # noqa
 from os_win import exceptions
 from os_win.utils import hostutils
 from os_win.utils.io import namedpipe
-from os_win.utils.storage.initiator import iscsi_cli_utils
-
-hyper_opts = [
-    cfg.BoolOpt('force_volumeutils_v1',
-                default=False,
-                help='DEPRECATED: Force V1 volume utility class',
-                deprecated_for_removal=True),
-]
-
-CONF = cfg.CONF
-CONF.register_opts(hyper_opts, 'hyperv')
 
 utils = hostutils.HostUtils()
 
@@ -50,11 +38,6 @@ utils_map = {
             'max_version': None,
             'path': 'os_win.utils.hostutils10.HostUtils10'}},
     'iscsi_initiator_utils': {
-        'ISCSIInitiatorCLIUtils': {
-            'min_version': 6.0,
-            'max_version': 6.2,
-            'path': 'os_win.utils.storage.initiator.iscsi_cli_utils.'
-                    'ISCSIInitiatorCLIUtils'},
         'ISCSIInitiatorUtils': {
             'min_version': 6.2,
             'max_version': None,
@@ -83,6 +66,11 @@ utils_map = {
             'min_version': 6.2,
             'max_version': None,
             'path': 'os_win.utils.metrics.metricsutils.MetricsUtils'}},
+    'migrationutils': {
+        'MigrationUtils': {
+            'min_version': 6.2,
+            'max_version': None,
+            'path': 'os_win.utils.compute.migrationutils.MigrationUtils'}},
     'networkutils': {
         'NetworkUtils': {
             'min_version': 6.2,
@@ -120,8 +108,12 @@ utils_map = {
     'vmutils': {
         'VMUtils': {
             'min_version': 6.2,
-            'max_version': 10,
+            'max_version': 6.3,
             'path': 'os_win.utils.compute.vmutils.VMUtils'},
+        'VMUtils6_3': {
+            'min_version': 6.3,
+            'max_version': 10,
+            'path': 'os_win.utils.compute.vmutils.VMUtils6_3'},
         'VMUtils10': {
             'min_version': 10,
             'max_version': None,
@@ -135,11 +127,16 @@ utils_map = {
         'DNSUtils': {
             'min_version': 6.2,
             'max_version': None,
-            'path': 'os_win.utils.dns.dnsutils.DNSUtils'}}
+            'path': 'os_win.utils.dns.dnsutils.DNSUtils'}},
+    'processutils': {
+        'ProcessUtils': {
+            'min_version': 6.2,
+            'max_version': None,
+            'path': 'os_win.utils.processutils.ProcessUtils'}}
 }
 
 
-def _get_class(class_type):
+def _get_class(class_type, *args, **kwargs):
     if class_type not in utils_map:
         raise exceptions.HyperVException(_('Class type %s does '
                                            'not exist') % class_type)
@@ -154,7 +151,8 @@ def _get_class(class_type):
         if (utils_class['min_version'] <= windows_version and
                 (utils_class['max_version'] is None or
                  windows_version < utils_class['max_version'])):
-            return importutils.import_object(utils_class['path'])
+            return importutils.import_object(utils_class['path'],
+                                             *args, **kwargs)
 
     raise exceptions.HyperVException(_('Could not find any %(class)s class for'
                                        'this Windows version: %(win_version)s')
@@ -163,7 +161,7 @@ def _get_class(class_type):
 
 
 def get_vmutils(host='.'):
-    return _get_class(class_type='vmutils')
+    return _get_class(class_type='vmutils', host=host)
 
 
 def get_vhdutils():
@@ -190,10 +188,7 @@ def get_pathutils():
     return _get_class(class_type='pathutils')
 
 
-def get_iscsi_initiator_utils(use_iscsi_cli=False):
-    use_iscsi_cli = use_iscsi_cli or CONF.hyperv.force_volumeutils_v1
-    if use_iscsi_cli:
-        return iscsi_cli_utils.ISCSIInitiatorCLIUtils()
+def get_iscsi_initiator_utils():
     return _get_class(class_type='iscsi_initiator_utils')
 
 
@@ -231,3 +226,11 @@ def get_clusterutils():
 
 def get_dnsutils():
     return _get_class(class_type='dnsutils')
+
+
+def get_migrationutils():
+    return _get_class(class_type='migrationutils')
+
+
+def get_processutils():
+    return _get_class(class_type='processutils')
